/*----------------------------------------------------------------------
  File    : distmat.c
  Contents: distance matrix management
  Author  : Christian Borgelt
  History : 2007.01.08 file created
            2007.01.10 first version completed
            2007.01.19 collection of largest distances added
            2007.01.21 function dm_nbgraph added
            2013.08.20 adapted to definition of type DIMID in matrix.h
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include "distmat.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
/* --- parameters for dv_tform, dm_tform --- */
const double dm_sqrt  [2] = { 0.5, 0 };
const double dm_square[2] = { 2.0, 0 };

/*----------------------------------------------------------------------
  Distance Vector Functions
----------------------------------------------------------------------*/

void dv_fromtab (double *dists, DIMID rowcnt, DIMID *ids, DIMID objcnt,
                 TABLE *tab, TUPLE *tpl)
{                               /* --- compute a distance vector */
  DIMID  i, k, n, e, o;         /* loop variables, buffer */
  double d;                     /* buffer for a distance */
  TUPLE  *t;                    /* to traverse the tuples */

  assert(dists && tab           /* check the function arguments */
  &&    (rowcnt >= 0) && (rowcnt <= tab_tplcnt(tab)));
  o = (objcnt < 0) ? -objcnt : objcnt;
  if (o >= rowcnt) objcnt = 0;  /* check objects against rows */
  if (objcnt == 0) {            /* if unrestricted number of objects */
    for (k = rowcnt; --k >= 0;) /* compute pairwise distances */
      dists[k] = tpl_sqrdist(tab_tpl(tab, k), tpl);
    return;                     /* fill the distance vector */
  }                             /* (ignore the object ids vector) */
  assert(ids);                  /* check the function arguments */
  dists[e = o-1] = (objcnt >= 0) ? INFINITY : -1;
  n = 0;                        /* init. the stored distances */
  for (k = rowcnt; --k >= 0; ) {
    t = tab_tpl(tab, k);        /* traverse the table, */
    if (t == tpl) continue;     /* but skip the tuple itself */
    d = tpl_sqrdist(t, tpl);    /* compute all pairwise distances */
    if (objcnt >= 0) {          /* if to collect smallest distances */
      if (d >= dists[e]) continue;
      for (i = n; (--i >= 0) && (d < dists[i]); ) {
        dists[i+1] = dists[i];  /* collect the objcnt smallest */
        ids  [i+1] = ids  [i];  /* distances in the matrix row */
      } }                       /* corresp. to the current tuple */
    else {                      /* if to collect largest distances */
      if (d <= dists[e]) continue;
      for (i = n; (--i >= 0) && (d > dists[i]); ) {
        dists[i+1] = dists[i];  /* collect the objcnt largest */
        ids  [i+1] = ids  [i];  /* distances in the matrix row */
      }                         /* corresp. to the current tuple */
    }
    dists[++i] = d;             /* note the identifiers of the */
    ids  [  i] = k;             /* objects with these distances */
    if (n < e) n++;             /* update the number of */
  }                             /* already collected distances */
}  /* dv_fromtab() */

/*--------------------------------------------------------------------*/

void dv_frommat (double *dists, DIMID rowcnt, DIMID *ids, DIMID objcnt,
                 MATRIX *mat, double *vec)
{                               /* --- compute a distance vector */
  DIMID  i, k, n, e, c, o;      /* loop variables, buffers */
  double d;                     /* buffer for a distance */
  double *v;                    /* to traverse the matrix rows */

  assert(dists && mat && vec    /* check the function arguments */
  &&    (rowcnt >= 0) && (rowcnt <= mat_rowcnt(mat)));
  o = (objcnt < 0) ? -objcnt : objcnt;
  if (o >= rowcnt) objcnt = 0;  /* check objects against rows */
  c = mat_colcnt(mat);          /* get the number of columns */
  if (objcnt == 0) {            /* if unrestricted number of objects */
    for (k = rowcnt; --k >= 0;) /* compute pairwise distances */
      dists[k] = vec_sqrdist(vec, mat_row(mat, k), c);
    return;                     /* fill the distance vector */
  }                             /* (ignore object ids vector) */
  assert(ids);                  /* check the function arguments */
  dists[e = o-1] = (objcnt >= 0) ? INFINITY : -1;
  n = 0;                        /* init. the stored distances */
  for (k = rowcnt; --k >= 0; ) {
    v = mat_row(mat, k);        /* traverse the matrix rows, */
    if (v == vec) continue;     /* but skip the vector itself */
    d = vec_sqrdist(vec, v, c); /* compute all pairwise distances */
    if (objcnt >= 0) {          /* if to collect smallest distances */
      if (d >= dists[e]) continue;
      for (i = n; (--i >= 0) && (d < dists[i]); ) {
        dists[i+1] = dists[i];  /* collect the objcnt smallest */
        ids  [i+1] = ids  [i];  /* distances in the matrix row */
      } }                       /* corresp. to the current tuple */
    else {                      /* if to collect largest distances */
      if (d <= dists[e]) continue;
      for (i = n; (--i >= 0) && (d > dists[i]); ) {
        dists[i+1] = dists[i];  /* collect the objcnt largest */
        ids  [i+1] = ids  [i];  /* distances in the matrix row */
      }                         /* corresp. to the current tuple */
    }
    dists[i] = d;               /* note the identifiers of the */
    ids  [i] = k;               /* objects with these distances */
    if (n < e) n++;             /* update the number of */
  }                             /* already collected distances */
}  /* dv_frommat() */

/*--------------------------------------------------------------------*/

void dv_tform (double *dists, DIMID cnt, const double *pars)
{                               /* --- transform distances */
  DIMID  i = 0;                 /* loop variable */
  double x = pars[0];           /* exponent of the transformation */
  double o = pars[1];           /* offset   of the transformation */

  assert(dists && pars);        /* check the function arguments */
  if (x < 0) { x = -x; i = cnt; }
  if      (x == 2.0) {          /* if to square distances */
    while (--cnt >= 0) dists[cnt] = dists[cnt] *dists[cnt] +o; }
  else if (x == 0.5) {          /* if to compute normal distances */
    while (--cnt >= 0) dists[cnt] = sqrt(dists[cnt])       +o; }
  else if (x != 1.0) {          /* if other exponent */
    while (--cnt >= 0) dists[cnt] = pow(dists[cnt], x)     +o;
  }                             /* transform the matrix elements */
  while (--i >= 0) dists[i] = exp(0.5 *dists[i]);
}  /* dv_tform() */

/*----------------------------------------------------------------------
  Distance Matrix Functions
----------------------------------------------------------------------*/

DISTMAT* dm_create (DIMID rowcnt, DIMID objcnt)
{                               /* --- create a distance matrix */
  DIMID   i, n, o;              /* loop variables */
  size_t  z;                    /* size of memory blocks */
  DISTMAT *dm;                  /* created distance matrix */
  double  *p;                   /* to traverse the matrix rows */
  DIMID   *q;                   /* to traverse the identifier vectors */

  assert(rowcnt > 0);           /* check the function arguments */
  z = (size_t)(rowcnt-1) *sizeof(double*);
  o = (objcnt < 0) ? -objcnt : objcnt;
  if (o >= rowcnt) objcnt = o = 0;   /* check objects against rows */
  if (objcnt != 0) {            /* if restricted number of objects */
    dm = (DISTMAT*)malloc(sizeof(DISTMAT) +z
                        +(size_t)rowcnt *sizeof(int*));
    if (!dm) return NULL;       /* create a distance matrix */
    z = (size_t)rowcnt *(size_t)(n = o);
    dm->ids = (DIMID**)(dm->dists +rowcnt);
    dm->dists[0] = p = (double*)calloc(z, sizeof(double));
    if (!p) {          free(dm); return NULL; }
    dm->ids  [0] = q = (DIMID*) calloc(z, sizeof(DIMID));
    if (!q) { free(q); free(dm); return NULL; }
    for (i = 0; ++i < rowcnt; ) {
      dm->ids  [i] = q += n;    /* allocate the matrix elements and */
      dm->dists[i] = p += n;    /* organize the matrix rows and */
    } }                         /* the object identifier vectors */
  else {                        /* if unrestricted number of objects */
    dm = (DISTMAT*)malloc(sizeof(DISTMAT) +z);
    if (!dm) return NULL;       /* create a triangular distance */
    dm->ids = NULL;             /* matrix without object identifiers */
    z = ((size_t)rowcnt *(size_t)(rowcnt-1)) /2;
    dm->dists[0] = p = (double*)calloc(z, sizeof(double));
    if (!p) { free(dm); return NULL; }
    for (i = 0; ++i < rowcnt; ) /* allocate the matrix elements */
      dm->dists[i] = p += i-1;  /* and organize the matrix rows */
  }                             /* (lower triangular matrix) */
  dm->rowcnt = rowcnt;          /* store the matrix parameters */
  dm->objcnt = objcnt;          /* (numbers of rows and objects) */
  return dm;                    /* return the created distance matrix */
}  /* dm_create() */

/*--------------------------------------------------------------------*/

void dm_delete (DISTMAT *dm)
{                               /* --- delete a distance matrix */
  assert(dm);                   /* check the function argument */
  if (dm->ids) free(dm->ids[0]);/* delete the object identifiers, */
  free(dm->dists[0]);           /* the distance matrix elements, */
  free(dm);                     /* and the base structure */
}  /* dm_delete() */

/*--------------------------------------------------------------------*/

double dm_getdist (DISTMAT *dm, DIMID rowid, DIMID idx, DIMID *id)
{                               /* --- get a distance */
  assert(dm                     /* check the function arguments */
      && (rowid >= 0) && (rowid < dm->rowcnt) && (idx >= 0));
  if (dm->objcnt != 0) {        /* if restricted number of objects */
    assert(idx < ((dm->objcnt < 0) ? -dm->objcnt : dm->objcnt));
    if (id) *id = dm->ids[rowid][idx];
    return dm->dists[rowid][idx]; } /* get object id and distance */
  else {                        /* if unrestricted number of objects */
    assert(idx < dm->rowcnt);   /* return the distance */
    if (idx < rowid) return dm->dists[rowid][idx];
    if (idx > rowid) return dm->dists[idx][rowid];
    return 0;                   /* make sure to access the */
  }                             /* lower triangle of the matrix */
}  /* dm_getdist() */

/*--------------------------------------------------------------------*/

void dm_setdist (DISTMAT *dm, DIMID rowid, DIMID idx, DIMID id,
                 double dist)
{                               /* --- set a distance */
  assert(dm                     /* check the function arguments */
      && (rowid >= 0) && (rowid < dm->rowcnt) && (idx  >= 0)
      && (id    >= 0) && (id    < dm->rowcnt) && (dist >= 0));
  if (dm->objcnt != 0) {        /* if restricted number of objects */
    assert(idx < ((dm->objcnt < 0) ? -dm->objcnt : dm->objcnt));
    dm->ids  [rowid][idx] = id; /* set object id and distance */
    dm->dists[rowid][idx] = dist; }
  else {                        /* if unrestricted number of objects */
    assert(idx < dm->rowcnt);   /* return the distance */
    if (idx < rowid) dm->dists[rowid][idx] = dist;
    if (idx > rowid) dm->dists[idx][rowid] = dist;
  }                             /* make sure to access the */
}  /* dm_getdist() */           /* lower triangle of the matrix */

/*--------------------------------------------------------------------*/

void dm_tform (DISTMAT *dm, const double *pars)
{                               /* --- transform distances in matrix */
  DIMID n = dm->rowcnt;         /* compute the number of distances */
  n = (dm->objcnt < 0) ? -dm->objcnt *n
    : (dm->objcnt > 0) ?  dm->objcnt *n : ((n *(n-1)) /2);
  dv_tform(dm->dists[0], n, pars);
}  /* dm_tform() */             /* transform the distances */

/*--------------------------------------------------------------------*/

void dm_fromtab (DISTMAT *dm, TABLE *tab)
{                               /* --- fill matrix from a table */
  DIMID i;                      /* loop variable */

  assert(dm && tab              /* check the function arguments */
     && (dm->rowcnt <= tab_tplcnt(tab)));
  if (dm->objcnt == 0) {        /* if unrestricted number of objects */
    for (i = dm->rowcnt; --i > 0; )
      dv_fromtab(dm->dists[i], i, NULL, 0, tab, tab_tpl(tab, i));
    return;                     /* traverse the tuples and */
  }                             /* compute triangular distance matrix */
  for (i = dm->rowcnt; --i >= 0; ) { /* traverse the tuples */
    dv_fromtab(dm->dists[i], dm->rowcnt, dm->ids[i], dm->objcnt,
               tab, tab_tpl(tab, i));
  }                             /* find distances and identifiers */
}  /* dm_fromtab() */           /* of the closests/farthest objects */

/*--------------------------------------------------------------------*/

void dm_frommat (DISTMAT *dm, MATRIX *mat)
{                               /* --- fill matrix from a matrix */
  DIMID i;                      /* loop variable */

  assert(dm && mat              /* check the function arguments */
     && (dm->rowcnt <= mat_rowcnt(mat)));
  if (dm->objcnt == 0) {        /* if unrestricted number of objects */
    for (i = dm->rowcnt; --i > 0; )
      dv_frommat(dm->dists[i], i, NULL, 0, mat, mat_row(mat, i));
    return;                     /* traverse the tuples and */
  }                             /* compute triangular distance matrix */
  for (i = dm->rowcnt; --i >= 0; ) { /* traverse the tuples */
    dv_frommat(dm->dists[i], dm->rowcnt, dm->ids[i], dm->objcnt,
               mat, mat_row(mat, i));
  }                             /* find distances and identifiers */
}  /* dm_frommat() */           /* of the closests/farthest objects */

/*--------------------------------------------------------------------*/

double dm_fromnbm (DISTMAT *dm, DISTMAT *nbm, double max)
{                               /* --- fill from neighorhood matrix */
  DIMID  i, j, k, o;            /* loop variables */
  DIMID  *ids;                  /* to traverse the identifier vectors */
  double *row;                  /* to traverse the distance elements */
  double d1, d2;                /* buffer for distances */

  assert(dm && nbm              /* check the function arguments */
  &&    (dm->rowcnt == nbm->rowcnt) && (nbm->objcnt > 0));
  for (i = dm->rowcnt; --i > 0; ) {
    row = dm->dists[i];         /* traverse all matrix elements */
    for (j = i; --j >= 0; ) row[j] = -1;
  }                             /* initialize to "no connection" */
  for (i = dm->rowcnt; --i >= 0; ) {
    row = nbm->dists[i];        /* traverse the given distance matrix */
    ids = nbm->ids  [i];        /* with the neighbor distances */
    for (j = nbm->objcnt; --j >= 0; ) {
      o = ids[j];               /* get the object identifier */
      if (o < i) dm->dists[i][o] = row[j];
      else       dm->dists[o][i] = row[j];
    }                           /* copy the given distances */
  }                             /* to the created matrix */
  for (k = dm->rowcnt; --k >= 0; ) {
    for (i = dm->rowcnt; --i > 0; ) {
      if (i == k) continue;     /* execute the Floyd-Warshall */
      row = dm->dists[i];       /* algorithm for finding */
      for (j = i; --j >= 0; ) { /* the smallest distances */
        if (j == k) continue;   /* between all pairs of objects */
        d1 = (k < i) ? dm->dists[i][k] : dm->dists[k][i];
        if (d1 < 0) continue;   /* get distance i -> k */
        d2 = (k < j) ? dm->dists[j][k] : dm->dists[k][j];
        if (d2 < 0) continue;   /* get distance k -> j */
        d1 += d2;               /* and sum the distances */
        if ((row[j] < 0) || (d1 < row[j])) row[j] = d1;
      }                         /* if a path through k leads */
    }                           /* to a shorter connection, */
  }                             /* replace the distance */
  if (max < 0) {                /* if factor for maximum given */
    for (d1 = 0, i = dm->rowcnt; --i > 0; ) {
      row = dm->dists[i];       /* traverse the non-negative entries */
      for (j = i; --j >= 0; )   /* of the created matrix */
        if (row[j] > d1) d1 = row[j];
    }                           /* determine the maximum distance */
    max *= -d1;                 /* and multiply this distance */
  }                             /* with the given factor */
  for (d1 = 0, i = dm->rowcnt; --i > 0; ) {
    row = dm->dists[i];         /* traverse the negative entries */
    for (j = i; --j >= 0; )     /* of the created matrix */
      if (row[j] < 0) row[j] = max;
  }                             /* set the maximum squared distance */
  return max;                   /* return the maximum distance */
}  /* dm_fromnbm() */

/*--------------------------------------------------------------------*/
#ifndef NDEBUG

void dm_show (DISTMAT *dm)
{                               /* --- show a distance matrix */
  DIMID   i, k, o;              /* loop variables */
  double *row;                  /* to traverse the matrix rows */
  DIMID  *ids;                  /* to traverse the identifier vectors */

  if (dm->objcnt == 0) {        /* if unrestricted number of objects */
    for (k = 0; k < dm->rowcnt; k++) {
      row = dm->dists[k];       /* traverse the matrix rows and */
      for (i = 0; i < k; i++)   /* the lower indices for each row */
        printf("%6.2f ", row[i]);
      printf("  0.00\n");       /* print the lower triangle */
    } }                         /* of the distance matrix */
  else {                        /* if restricted neighborhood */
    for (k = 0; k < dm->rowcnt; k++) {
      row = dm->dists[k];       /* traverse the matrix rows and */
      ids = dm->ids  [k];       /* get distances and object ids */
      o = (dm->objcnt < 0) ? -dm->objcnt : dm->objcnt;
      for (i = 0; i < o; i++)   /* traverse the objects */
        printf("%2"DIMID_FMT":%6.2f ", ids[i], row[i]);
      printf("\n");             /* print an index/distance pair */
    }                           /* for each matrix element */
  }
}  /* dm_show() */

#endif
