/*----------------------------------------------------------------------
  File    : ptli.c
  Contents: prototype-less cluster induction
  Authors : Christian Borgelt
  History : 2007.01.11 file created
            2007.01.19 first version completed
            2007.01.21 use of a neighborhood graph added
            2007.01.23 stored number of objects to consider
            2007.02.27 offset for membership degrees added
            2007.10.10 evaluation of attribute directions added
            2013.08.20 adapted to definitions ATTID, VALID, TPLID etc.
            2014.10.24 changed from LGPL license to MIT license
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <time.h>
#include <assert.h>
#include "params.h"
#ifndef AS_READ
#define AS_READ
#endif
#ifndef AS_PARSE
#define AS_PARSE
#endif
#ifndef AS_DESC
#define AS_DESC
#endif
#include "attset.h"
#ifndef TAB_READ
#define TAB_READ
#endif
#include "table.h"
#include "ptlset.h"
#include "random.h"
#include "error.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PRGNAME     "ptli"
#define DESCRIPTION "prototype-less cluster set induction"
#define VERSION     "version 2.1 (2014.10.24)         " \
                    "(c) 2007-2014   Christian Borgelt"

/* --- error codes --- */
/* error codes 0 to -5 defined in attset.h */
#define E_OPTION     (-6)       /* unknown option */
#define E_OPTARG     (-7)       /* missing option argument */
#define E_ARGCNT     (-8)       /* wrong number of arguments */
#define E_PARSE      (-9)       /* parse errors on input file */
#define E_TPLCNT    (-10)       /* data file is empty */
#define E_ATTCNT    (-11)       /* no usable attributes found */
#define E_MATCH     (-12)       /* mismatch of number of rows */
#define E_MSEXP     (-13)       /* invalid membership exponent */
#define E_MSOFF     (-14)       /* invalid membership offset */

#define SEC_SINCE(t)  ((double)(clock()-(t)) /(double)CLOCKS_PER_SEC)

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
static const char *errmsgs[] = {   /* error messages */
  /* E_NONE      0 */  "no error",
  /* E_NOMEM    -1 */  "not enough memory",
  /* E_FOPEN    -2 */  "cannot open file %s",
  /* E_FREAD    -3 */  "read error on file %s",
  /* E_FWRITE   -4 */  "write error on file %s",
  /* E_STDIN    -5 */  "double assignment of standard input",
  /* E_OPTION   -6 */  "unknown option -%c",
  /* E_OPTARG   -7 */  "missing option argument",
  /* E_ARGCNT   -8 */  "wrong number of arguments",
  /* E_PARSE    -9 */  "parse error(s) on file %s",
  /* E_TPLCNT  -10 */  "data file is empty",
  /* E_ATTCNT  -11 */  "no (usable) attributes (need at least 1)",
  /* E_MATCH   -12 */  "mismatch in number of tuples "
                         "(%"DIMID_FMT" vs. %"DIMID_FMT")",
  /* E_MSEXP   -13 */  "invalid membership exponent %g",
  /* E_MSOFF   -14 */  "invalid membership offset %g",
  /*           -15 */  "unknown error"
};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static CCHAR   *prgname;        /* program name for error messages */
static SCANNER *scan    = NULL; /* scanner (for domain descriptions) */
static ATTSET  *attset  = NULL; /* attribute set */
static TABREAD *tread   = NULL; /* table reader */
static TABLE   *table   = NULL; /* table  of training tuples */
static DISTMAT *distmat = NULL; /* distance matrix */
static PTLSET  *ptlset  = NULL; /* prototype-less cluster set */
static FILE    *out     = NULL; /* output file */

/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/

#ifndef NDEBUG
  #undef  CLEANUP               /* clean up memory and close files */
  #define CLEANUP \
  if (ptlset)  ptl_delete(ptlset);   \
  if (distmat) dm_delete(distmat);   \
  if (table)   tab_delete(table, 0); \
  if (attset)  as_delete(attset);    \
  if (tread)   trd_delete(tread, 1); \
  if (scan)    scn_delete(scan,  1); \
  if (out && (out != stdout)) fclose(out);
#endif

GENERROR(error, exit)           /* generic error reporting function */

/*--------------------------------------------------------------------*/

int main (int argc, char* argv[])
{                               /* --- main function */
  int     i, k = 0;             /* loop variables, counters */
  char    *s;                   /* to traverse options */
  CCHAR   **optarg = NULL;      /* option argument */
  CCHAR   *fn_dom  = NULL;      /* name of domain file */
  CCHAR   *fn_hdr  = NULL;      /* name of table header file */
  CCHAR   *fn_tab  = NULL;      /* name of table file */
  CCHAR   *fn_cls  = NULL;      /* name of output file */
  CCHAR   *blanks  = NULL;      /* blank  characters */
  CCHAR   *fldseps = NULL;      /* field  separators */
  CCHAR   *recseps = NULL;      /* record separators */
  CCHAR   *comment = NULL;      /* comment characters */
  int     mode     = AS_ATT|AS_NOXATT|AS_NONULL; /* table read mode */
  VALID   clscnt   = 2;         /* number of clusters */
  DIMID   objcnt   = 0;         /* number of objects to consider */
  int     owrite   = 0;         /* overwrite parameters with options */
  int     norm     = 1;         /* flag for input range normalization */
  int     inimode  = PTL_RANDOM;/* cluster initialization mode */
  int     regwgt   = 0;         /* regularize weight distribution */
  DIMID   epochs   = 1000, e;   /* number of update epochs */
  double  ngmax    = 0;         /* limit for a neighborhood graph */
  double  dtps[]   = {2, 0};    /* distance transformation parameters */
  double  msexp    = 2;         /* membership exponent for adaptation */
  double  msoff    = 0;         /* membership offset   for adaptation */
  double  tfexp    = 1;         /* exponent for sum transformation */
  double  trmchg   = 0;         /* maximum change for termination */
  int     maxlen   = 0;         /* maximal output line length */
  long    seed;                 /* seed value for random numbers */
  double  chg;                  /* maximal change of a m.s. degree */
  double  obj;                  /* value of the objective function */
  DISTMAT *ngdm;                /* neighborhood graph distance matrix */
  ATTID   m, c;                 /* number of attributes */
  TPLID   n, r;                 /* number of data tuples */
  double  w;                    /* weight of data tuples */
  clock_t t;                    /* timer for measurements */

  prgname = argv[0];            /* get program name for error msgs. */
  seed    = (long)time(NULL);   /* and get a default seed value */

  /* --- print startup/usage message --- */
  if (argc > 1) {               /* if arguments are given */
    fprintf(stderr, "%s - %s\n", argv[0], DESCRIPTION);
    fprintf(stderr, VERSION); } /* print a startup message */
  else {                        /* if no argument is given */
    printf("usage: %s [options] domfile "
                     "[-d|-h hdrfile] tabfile clsfile\n", argv[0]);
    printf("%s\n", DESCRIPTION);
    printf("%s\n", VERSION);
    printf("-c#      number of clusters                     "
                    "(default: %"VALID_FMT")\n", clscnt);
    printf("-k#      number of objects to consider          "
                    "(default: all)\n");
    printf("-q       do not normalize input ranges          "
                    "(default: mean=0, var=1)\n");
    printf("-p#:#    distance transformation parameters     "
                    "(default: %g:%g)\n", dtps[0], dtps[1]);
    printf("-g#      neighborhood graph limit               "
                    "(default: distances directly)\n");
    printf("-S#      seed for random numbers                "
                    "(default: time)\n");
    printf("-i       initialize by selecting seeds          "
                    "(default: random memberships)\n");
    printf("-x#      exponent for membership degrees        "
                    "(default: %g)\n", msexp);
    printf("-y#      offset   for membership degrees        "
                    "(default: %g)\n", msoff);
    printf("-X#      exponent for term transformation       "
                    "(default: %g)\n", tfexp);
    printf("-O       overwrite with option values           "
                    "(default: parameters read)\n");
    printf("-R       regularize weight distribution         "
                    "(default: keep distribution)\n");
    printf("-e#      maximum number of update epochs        "
                    "(default: %"DIMID_FMT")\n", epochs);
    printf("-T#      maximum change for termination         "
                    "(default: %g)\n", trmchg);
    printf("-l#      output line length                     "
                    "(default: no limit)\n");
    printf("-b#      blank   characters                     "
                    "(default: \" \\t\\r\")\n");
    printf("-f#      field   separators                     "
                    "(default: \" \\t,\")\n");
    printf("-r#      record  separators                     "
                    "(default: \"\\n\")\n");
    printf("-C#      comment characters                     "
                    "(default: \"#\")\n");
    printf("-n       number of tuple occurrences in last field\n");
    printf("domfile  file containing domain descriptions\n");
    printf("-d       use default header "
                    "(field names = field numbers)\n");
    printf("-h       read table header (field names) from hdrfile\n");
    printf("hdrfile  file containing table header (field names)\n");
    printf("tabfile  table file to read "
                    "(field names in first record)\n");
    printf("outfile  file to write induced clustering to\n");
    return 0;                   /* print a usage message */
  }                             /* and abort the program */

  /* --- evaluate arguments --- */
  for (i = 1; i < argc; i++) {  /* traverse arguments */
    s = argv[i];                /* get option argument */
    if (optarg) { *optarg = s; optarg = NULL; continue; }
    if ((*s == '-') && *++s) {  /* -- if argument is an option */
      while (1) {               /* traverse characters */
        switch (*s++) {         /* evaluate option */
          case 'c': clscnt  = (int)strtol(s, &s, 0); break;
          case 'k': objcnt  = (int)strtol(s, &s, 0); break;
          case 'q': norm    = 0;                     break;
          case 'g': ngmax   =      strtod(s, &s);    break;
          case 'S': seed    =      strtol(s, &s, 0); break;
          case 'i': inimode = PTL_SEED|PTL_UNIFORM;  break;
          case 'R': regwgt  = 1;                     break;
          case 'p': getdblvec(s, &s, 2, dtps);       break;
          case 'x': msexp   =      strtod(s, &s);    break;
          case 'y': msoff   =      strtod(s, &s);    break;
          case 'X': tfexp   =      strtod(s, &s);    break;
          case 'O': owrite  = 1;                     break;
          case 'e': epochs  = (int)strtol(s, &s, 0); break;
          case 'T': trmchg  =      strtod(s, &s);    break;
          case 'l': maxlen  = (int)strtol(s, &s, 0); break;
          case 'b': optarg  = &blanks;               break;
          case 'f': optarg  = &fldseps;              break;
          case 'r': optarg  = &recseps;              break;
          case 'C': optarg  = &comment;              break;
          case 'n': mode   |= AS_WEIGHT;             break;
          case 'd': mode   |= AS_DFLT;               break;
          case 'h': optarg  = &fn_hdr;               break;
          default : error(E_OPTION, *--s);           break;
        }                       /* set option variables */
        if (!*s) break;         /* if at end of string, abort loop */
        if (optarg) { *optarg = s; optarg = NULL; break; }
      } }                       /* get option argument */
    else {                      /* if argument is no option */
      switch (k++) {            /* evaluate non-option */
        case  0: fn_dom = s;      break;
        case  1: fn_tab = s;      break;
        case  2: fn_cls = s;      break;
        default: error(E_ARGCNT); break;
      }                         /* note filenames */
    }
  }
  if (optarg) error(E_OPTARG);  /* check option argument */
  if (k != 3) error(E_ARGCNT);  /* check number of arguments */
  if (fn_hdr && (strcmp(fn_hdr, "-") == 0))
    fn_hdr = "";                /* convert "-" to "" */
  i = (!fn_dom || !*fn_dom) ? 1 : 0;
  if  (!fn_tab || !*fn_tab) i++;
  if  ( fn_hdr && !*fn_hdr) i++;/* check assignments of stdin: */
  if (i > 1) error(E_STDIN);    /* stdin must not be used twice */
  if (fn_hdr)                   /* set header flags */
    mode = AS_ATT | (mode & ~AS_DFLT);
  if (msexp <=  0) error(E_MSEXP, msexp);
  if (msoff <= -1) error(E_MSOFF, msoff);
  if (tfexp <   0) tfexp = 1.0/(msexp -1.0);
  rseed((unsigned int)seed);    /* init. the random number generator */
  fputc('\n', stderr);          /* terminate the startup message */

  /* --- parse domain descriptions --- */
  attset = as_create("domains", att_delete);
  if (!attset) error(E_NOMEM);  /* create an attribute set */
  scan = scn_create();          /* create a scanner */
  if (!scan)   error(E_NOMEM);  /* for the domain file */
  t = clock();                  /* start timer, open input file */
  if (scn_open(scan, NULL, fn_dom) != 0)
    error(E_FOPEN, scn_name(scan));
  fprintf(stderr, "reading %s ... ", scn_name(scan));
  if (as_parse(attset, scan, AT_ALL, 1) != 0)
    error(E_PARSE, scn_name(scan));
  m = as_attcnt(attset);        /* parse domain descriptions */
  as_target(attset, NULL, 0);   /* reduce to input attributes */
  if (as_attcnt(attset) <= 0)   /* check for at least one attribute */
    error(E_ATTCNT);            /* that can be used for training */

  /* --- read cluster descriptions --- */
  if (scn_eof(scan, 0))         /* if there is no cluster set */
    fprintf(stderr, "[%"ATTID_FMT" attribute(s)]", m);
  else {                        /* if there is a cluster set */
    ptlset = ptl_parse(scan);   /* parse the cluster set */
    if (!ptlset || !scn_eof(scan, 1)) error(E_PARSE, scn_name(scan));
    if (!owrite)                /* if to keep the parameters */
      objcnt = ptl_getobjc(ptlset);
    else {                      /* if to overwrite the parameters */
      ptl_msexp(ptlset, msexp, msoff);
      ptl_tfexp(ptlset, tfexp); /* set exponents and offset */
      ptl_setobjc(ptlset, objcnt);
    }                           /* set the number of objects */
    fprintf(stderr, "[%"ATTID_FMT" attribute(s),", m);
    fprintf(stderr, " %"DIMID_FMT" cluster(s)]", ptl_clscnt(ptlset));
  }                             /* print a success message */
  scn_delete(scan, 1);          /* delete the scanner and */
  scan = NULL;                  /* clear the scanner variable */
  fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));

  /* --- read table header --- */
  tread = trd_create();         /* create a table reader and */
  if (!tread) error(E_NOMEM);   /* configure the characters */
  trd_allchs(tread, recseps, fldseps, blanks, "", comment);
  if (fn_hdr) {                 /* if a header file is given */
    t = clock();                /* start timer, open input file */
    if (trd_open(tread, NULL, fn_hdr) != 0)
      error(E_FOPEN, trd_name(tread));
    fprintf(stderr, "reading %s ... ", trd_name(tread));
    k = as_read(attset, tread, (mode & ~AS_DFLT) | AS_ATT);
    if (k < 0) error(-k, as_errmsg(attset, NULL, 0));
    trd_close(tread);           /* read table header, close file */
    m = as_attcnt(attset);      /* get the number of attributes */
    fprintf(stderr, "[%"ATTID_FMT" attribute(s)]", m);
    fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
    mode &= ~(AS_ATT|AS_DFLT);  /* print a success message and */
  }                             /* remove the attribute flag */

  /* --- read table --- */
  t = clock();                  /* start timer, open input file */
  if (trd_open(tread, NULL, fn_tab) != 0)
    error(E_FOPEN, trd_name(tread));
  fprintf(stderr, "reading %s ... ", trd_name(tread));
  table = tab_create("table", attset, tpl_delete);
  if (!table) error(E_NOMEM);   /* create a data table */
  k = tab_read(table, tread, mode);
  if (k < 0) error(-k, tab_errmsg(table, NULL, 0));
  trd_delete(tread, 1);         /* read the table body and */
  tread = NULL;                 /* delete the table reader */
  m = tab_attcnt(table);        /* get the number of attributes */
  n = tab_tplcnt(table);        /* and the number of data tuples */
  w = tab_tplwgt(table);        /* and print a success message */
  fprintf(stderr, "[%"ATTID_FMT" attribute(s),", m);
  fprintf(stderr, " %"TPLID_FMT, n);
  if (w != (double)n) fprintf(stderr, "/%g", w);
  fprintf(stderr, " tuple(s)] done [%.2fs].\n", SEC_SINCE(t));
  if (n <= 0) error(E_TPLCNT);  /* check for at least one tuple */

  /* --- create cluster set --- */
  if (ptlset) {                 /* if a cluster set was read */
    if (n != ptl_rowcnt(ptlset))/* check the number of tuples */
      error(E_MATCH, ptl_rowcnt(ptlset), n); }
  else {                        /* if no cluster set was read */
    t = clock();                /* start timer, print message */
    fprintf(stderr, "creating cluster set ... ");
    ptlset = ptl_create(n, clscnt);
    if (!ptlset) error(E_NOMEM);/* create a cluster set */
    ptl_setobjc(ptlset, objcnt);/* set number of objects to consider */
    ptl_setnorm(ptlset, norm);  /* and the normalization flag */
    ptl_msexp(ptlset, msexp, msoff);
    ptl_tfexp(ptlset, tfexp);   /* set exponents and offset */
    ptl_init(ptlset, inimode, drand);
    if (norm) {                 /* if to normalize input ranges */
      for (c = 0; c < m; c++) { /* traverse the table columns */
        k = tab_coltype(table, c);
        if      (k == AT_INT) tab_colconv(table, c, AT_FLT);
        else if (k != AT_FLT) continue;
        tab_colnorm(table, c, 0, 1, NULL, NULL);
      }                         /* normalize all numeric columns */
    }                           /* to mean 0 and std. deviation 1 */
    fprintf(stderr, "[%"TPLID_FMT" tuple(s),",   n);
    fprintf(stderr, " %"DIMID_FMT" cluster(s)]", clscnt);
    fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
  }                           /* print a success message */
  if (mode & AS_WEIGHT)         /* if there are tuple weights, */
    for (r = 0; r < n; r++)     /* set the object weights */
      ptl_setwgt(ptlset, r, tpl_getwgt(tab_tpl(table, r)));

  /* --- create distance matrix --- */
  t = clock();                  /* start the timer */
  fprintf(stderr, "computing distances ... ");
  distmat = dm_create(n, objcnt);
  if (!distmat) error(E_NOMEM); /* create a distance matrix */
  dm_fromtab(distmat, table);   /* and fill it from the table */
  if ((ngmax != 0)              /* if to use a neighborhood graph */
  &&  (objcnt > 0)) {           /* and a neighborhood is given */
    ngdm = dm_create(n, 0);     /* create another distance matrix */
    if (!ngdm) error(E_NOMEM);  /* (with full distance coverage) */
    dm_tform(distmat, dm_sqrt); /* compute actual distances */
    dm_fromnbm(ngdm, distmat, ngmax);
    dm_tform(ngdm, dm_square);  /* fill the new distance matrix */
    dm_delete(distmat);         /* with neighborhood graph distances, */
    distmat = ngdm;             /* square the distances, and */
  }                             /* replace the original matrix */
  dtps[0] /= 2;                 /* adapt transformation parameters */
  dm_tform(distmat, dtps);      /* and transform the distances */
  fprintf(stderr, "[%"TPLID_FMT" tuple(s)]", n);
  fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));

  /* --- induce prototype-less cluster set --- */
  t = clock();                  /* start the timer */
  fprintf(stderr, "inducing clusters ... ");
  for (e = 0; e < epochs; e++){ /* compute a maximum number of epochs */
    if ((e & 0x00ff) == 0)      /* print the current number of epochs */
      fprintf(stderr, "%8"DIMID_FMT"\b\b\b\b\b\b\b\b", e);
    chg = ptl_update(ptlset, distmat, NULL, drand);
    if (chg < 0) error(E_NOMEM);/* update the cluster set */
    if (regwgt) ptl_regular(ptlset);
    if (chg <= trmchg) break;   /* and check for termination */
  }                             /* print a success message */
  obj = ptl_objfn(ptlset, distmat);
  fprintf(stderr, "[%"DIMID_FMT" epoch(s): %g]", e, obj);
  fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));

  /* --- write clustering result --- */
  t = clock();                  /* start timer, open output file */
  if (fn_cls && (strcmp(fn_cls, "-") == 0)) fn_cls = "";
  if (fn_cls && *fn_cls) { out = fopen(fn_cls, "w"); }
  else                   { out = stdout; fn_cls = "<stdout>"; }
  fprintf(stderr, "writing %s ... ", fn_cls);
  if (as_desc(attset, out, AS_TITLE|AS_IVALS, maxlen) != 0)
    error(E_FWRITE, fn_cls);    /* describe attribute domains */
  fprintf(out, "\n");           /* leave one line empty */
  if (ptl_desc(ptlset, out, PTL_TITLE|PTL_INFO, maxlen) != 0)
    error(E_FWRITE, fn_cls);    /* describe the cluster set */
  if (fn_tab && *fn_tab)        /* record the data table */
    fprintf(out, "/* %s */\n", fn_tab);
  if (out && (((out == stdout) ? fflush(out) : fclose(out)) != 0))
    error(E_FWRITE, fn_cls);    /* close the output file and */
  out = NULL;                   /* print a success message */
  fprintf(stderr, "[%"TPLID_FMT" tuple(s),",   n);
  fprintf(stderr, " %"DIMID_FMT" cluster(s)]", clscnt);
  fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));

  /* --- clean up --- */
  CLEANUP;                      /* clean up memory and close files */
  SHOWMEM;                      /* show (final) memory usage */
  return 0;                     /* return 'ok' */
}  /* main() */
